# Frame-Recurrent Video Denoising

Code for training and testing EMVD (Efficient Multi-Stage Video Denoising) method. 
It is a recurrent method in which one intermediate feature is taken as input for processing next frame.

To use this type of recurrence, one must specify **models** in the code:  `recurrentFeat`, which is for feature recurrent nets.

This model can be used with different **network architectures**. In our case, the `EMVD` is exactly what we use.

## Requirements

The code is mostly python with some C code as well.

It requires the following python packages:
```bash
kornia
torch==1.8.0+cu111
torchvision==0.9.0+cu111
numpy
opencv-contrib-python
scipy
scikit-image
matplotlib
Pillow
imagecodecs
```

And the `fftw` library, for the compilation of the C code. For compiling the
code use:
```bash
mkdir -p build && cd build && cmake .. && make
```
Note: If the fftw external library is not present in your system please set the
flag `useFFTW` to `OFF` in [CMakeLists.txt](./CMakeLists.txt).


## Usage

Use `train.py` to train the model. For testing, we use the snippet of validation part to generate results currently.



### Dataset generation

Before training, we need a dataset. To make use of the code, the dataset should contain a series of frame sequences, arranged with the following folder structure:  
```
dataroot/
 ├── raw/                            # original .raw images
 │   ├── 000/
 │   ├── 001/
 │   ...
 ├── gt/                             # .raw converted to tiff
 │   ├── 000/
 │   ├── 001/
 │   ...
 ├── noisy                           # added noise of specified ISO
 │   ├── 000/
 │   ├── 001/
 │   ...
 ├── flow                            # pre-computed flow between noisy frames
 │   └── noisy
 │       └── tvl1_backflow
 │           └── noisyinputs
 │               ├── 000/
 │               ├── 001/
 │               ...
 ├── warped                          # warped noisy frames
 │   └── noisy
 │       └── tvl1_backflow
 │           └── noisyinputs
 │               ├── 000/
 │               ├── 001/
 │               ...
 └── masked                          # warped and masked noisy frames
     └── noisy
         └── tvl1_backflow
             └── noisyinputs
                 ├── 000/
                 ├── 001/
                 ...
```
The `000/, 001/` are folders associated to each sequence. The images, flows or
masks are stored as `.tiff` files (or .raw files in the case of the raw data).

For RAW data, the dataset is generated automatically in the code. The only folder required 
is the `raw/` folder, and the code will generate the rest. Examples will be shown below.

### Dataloaders

Before going to the usage examples we need to talk about the dataloaders. The
dataloaders read the data from the hard-drive and pre-process it to feed the
model. The training dataloader operates differently than testing ones.

**Training dataloaders:** Training dataloaders sample 3D patches at random
position from the videos in the training set. To speed up the this operation,
they work as follows:
  1. At the beginning of each epoch, a number of consecutive frames is read
	  from each video in the training set. The first frames of these segments is
	  chosen at random.
	  This number is called `frames2load`.
	  Thus, if there are `N` videos in the training set, we will load into the
	  CPU RAM `N * frames2load` frames. To speed up training, we pre-compute
	  as many things as possible, such as optical flows warping, etc.
	  For each frame it loads the necessary
	  data for running the network and computing the loss, i.e. noisy frames, clean frames, and optical flows, etc.
  2. From these `N` segments of video, we extract a number of 3D smaller
	  crops at random positions (crops may overlap). The number 
	  of crops is called `max-dataset-size`. These crops are fed to the
	  network in a random order, and the epoch ends after all of them
	  are processed.

**Testing dataloaders:** During testing the dataloader reproduces the
use case of denoising a single video sequence. It processes the
videos sequentially loading whole frames (not crops).

[comment]: <> (`axel`: *dataloader for training non-recurrent networks.* The inputs for the non-recurrent networks are the noisy frames, which are known beforehand. When the training is launched with warping, the dataloader checks for the existence of necessary warped noisy frames. If they don't exists, it computes the optical flow and warp them. The same applies for the mask. These warped frames are stored on the hard drive in the dataset. So the next time they are  equired they won't be recomputed.)

The `--dataset_mode` should be specified as `axel4rec` and  `infer4rec` for training and test process respectively. 
In this case the network requires as input its own previous outputs. Therefore, the warping cannot
be precomputed offline. Ideally, we should compute the optical flows and the warpings
using the previous network outputs.

In practice we can't train like this, as it would be too slow and unstable
(at the beginning of training the network produces pure random noise that
 can't be matched with an optical flow).
Instead, we compute
the optical flow using the noisy images `noisy[t]` and `noisy[t-1]`.
In this way the can be pre-computed off-line. During training the dataloader
provides them and we only need to apply the warping.

```
dataroot/       # axel  |  axel4rec | inferaxel
 ├── raw/       #       |           |
 ├── gt/        #  x    |    x      |   x  
 ├── noisy      #  x    |    x      |   x  
 ├── flow       #       |    x      |
 ├── warped     #  x    |           |   x
 └── masked     #  x    |           |   x

Table: which data is provided by each dataloader.
```

The output of each dataloader are dictionaries. For example, for a recurrent model, axel4rec provides:

```
     'gt' - tensor of shape [B,C,H,W] with groundtruth frames.
      'n' - tensor of shape [B,C,H,W] with Noisy frames.
   'flow' - tensor of shape [B,TD,D,2,H,W] with optical flows from D previous frames to the target frame from each unrolling. 
'gt_path' - string with paths to the first groundtruth frames of the batch.
 'n_path' - string with paths to the first noisy frames of the batch.
```


Remark: `H` and `W` denote the height and width of each frame (might be cropped with respect to the original) to be fed to the network. In training, `H=W` since we use square crops. `C` is the number of concatenated channels from each image needed to do the denoising of the current frame, i.e. `4*(TD+D)` for raw images. `B` is the batch size. `N` is the number of frames (including the current one) needed to denoised the current frame. Finally, `D` denotes the number of previous frames and `TD` the number of unrollings. For more details, refer to `recurrentFeat_model.py` comments.



### Training

The input of the training is a dataset, and a number of parameters controlling the dataloader, the model and the training itself. The output is a series of **checkpoints** files, which store the status of training 
(including the network weights) at the end of each epoch.

The `train.py` function receives many command line arguments. 
We will demonstrate some of the most commonly used ones with a series of examples.
To list all options type:
```
python3 train.py --help
```



**EXAMPLE. Training a recurrent model.**  Training a recurrent network needs to perform during training several iterations of the recurrency, which is called unrollings. If we want to perform
4 unrollings, we need to load crops with 5 frames:
```python
deno[1] = net(noisy[0]-warped-to-1, noisy[1])
deno[2] = net( deno[1]-warped-to-2, noisy[2])
deno[3] = net( deno[2]-warped-to-3, noisy[3])
deno[4] = net( deno[3]-warped-to-4, noisy[4])
```
The loss will compute a weighted combination of the loss for the output of each unrolling:
```python
total_loss = weight[1] * loss(deno[1], gt[1])
           + weight[2] * loss(deno[2], gt[2])
           + weight[2] * loss(deno[3], gt[3])
           + weight[2] * loss(deno[4], gt[4])
```
Remark: Above gives a general description of exploiting recurrency. When feature is used as recurrency, the *deno* can be replaced accordingly.

Thus for training recurrent networks we need to specify:
- number of frames that are inputed to the network: `model_patch_depth`
- number of frames loaded by the dataloader `patch_depth`
- weighting of the individual losses in the total loss: `unroll_focus`

```bash
python3 train.py \
    --model recurrentFeat \                  # because it's a recurrent model
    --model_patch_depth 2 \                  # number of frames the inputed to the network
    --warp \                                 # warp previous frames to align them to current frame
    --netDenoiser EMVDfeatbasic-VST=True-ISO=3200 \ # specify network architecture
    --dataroot /path/to/train_dataset \      # path to root folder of dataset
    --val_dataroot /path/to/val_dataset      # path to root folder of val dataset
    --data_nocheck \                         # skip checking if dataset is correct
    --dataset_mode axel4rec \                # dataloader for recurrent networks
    --max_dataset_size 90000 \               # number of crops processed per epoch
    --niter 70  \                            # 70 epochs with the initial learning rate (lr) ...
    --niter_decay 30 \                       # ... plus 30 with a linear reduction of the lr
    --lr 0.00005 --beta1 0.9 \               # initial learning rate and Adam momentum parameter
    --batch_size 2 \                         # batch size of 2
    --patch_width 136 \                      # crops of spatial size 272x272
    --patch_depth 5 \                        # dataloader will load crops 272x272x5 (4 unrollings)
    --norm none  \                           # no normalization layers (e.g. batch norm)
    --no_dropout \                           # no dropout
    --unroll_focus gradual04_from20 \        # dynamic weighting in the loss (more on this later)
    --visuals none \                         # don't generate visual results during training
    --val_visuals denoiser  \                # generate visual results during validation 
    --gpu_ids 0 \                            # use gpu 0
    --suffix 4unrolls-bs2-p136-trial1 \           # suffix for the checkpoints folder name
    --checkpoints_dir /path/to/checkpoints \ # where to place the folder with checkpoints
    --autoresume                             # resume from the latest previous checkpoint
    --warp  \                                # adopt optical flow to warp
    --no_frame_rec \                         # feature recurrency, not frame recurrency
    --nFolder noisy_iso3200 --gtFolder gt_iso3200 --rawFolder raw \ # folder of noisy, ground truth, and raw
    --lambda_pre_L1 1                        # weight for loss of pre-denoising
```
Remark: the first thing done by the training function is to check that all necessary data exists. If
a file doesn't exist, it is computed and stored according to the folder structure described before.
The `--ISOgen` flag tells the script which is the noise level that will be added to the 
clean raw images. It will be ignored if noisy images are found in the `noisy` folder.
The generation of the dataset can take hours.

The training code will create a folder in the `checkponts_dir` where the produced checkpoints
(and other log files will be stored). The name of the folder consists of an
automatically generated prefix, and a manually chosen suffix:
```
/path/to/checkpoints/recurrentFeat-EMVDfeatbasic-VST=True-ISO=3200-noisyinputs-warp-raw-i4o4-4unrolls-bs2-p136-trial1/
|-checkpoints_dir---|---output folder prefix ----------------|--- suffix ----------|
```
These are the contents of the output folder after 32 epochs of training
```
1_net_Denoise.pth           # checkpoint after epoch 1
2_net_Denoise.pth
...
32_net_Denoise.pth          # checkpoint after epoch 32
latest_net_Denoise.pth      # latest checkpoint saved
loss_log.txt                # log of the training losses
opt_train.txt               # log of all the options of the training
status.pkl                  # dictionary with training status (epoch, lr, etc) 
                            # at the last completed epoch. Used by autoresume.
```

**More on the unroll focus.** These are the available options for the
weights in the loss, with examples for 6 unrollings.
- `all`:     uniform weights                                `[1/6, 1/6, 1/6, 1/6, 1/6, 1/6]`
- `last`:    only last unrolling                            `[ 0 ,  0 ,  0 ,  0 ,  0 ,  1 ]`
- `ge[ii]`:  only unrolling greater or equal than ii. e.g.  `ge3 = [ 0 ,  0 ,  0 , 1/3, 1/3, 1/3]`
- `inc[ee]`: weight for unrolling iteration `u` is prop to `u**ee`. e.g. `inc2 = [ 1 ,  4 ,  9 , 16 , 25 , 36 ]/91`
- `graduni[ii]_from[jj]`: gradual transition from non-recurrent to uniform, between epochs `jj` and `ii+jj`
- `gradual[ii]_from[jj]`: gradual transition from non-recurrent to 90% of weight in the last unrolling, between epochs `jj` and `ii+jj`

More documentation about the gradual weights is available in the comments in the code.

### Testing

This examples show how to compute results on the validation set. As said before, we use the validation part of train.py to generate results. 
```bash
# non-recurrent model. compute outputs over three videos of the validation set using the 
# weights at the end of epoch 86
python3 train.py \                             # same script as training
  --path2epoch /path-of-checkpoints/86 \                                # load the weights at the end of epoch 86
  ......................................
  --suffix 4unrolls-bs2-p136-validation \            # the suffix
  --checkpoints_dir /path/to/checkpoints \    # root of checkpoint folders
  --lr 0 \                                   # set lr 0
  --val_visuals denoised \                        # store the denoised images
```

### Post-processing
```bash
python3 ./fwd_pipeline/post_processing.py  \
			--input path-to-raw/00000%03d_denoised.tif \
			--output path-to-RGB/00000%03d_denoised.png \
			--first 003 --last 267 --step 3	--seq 000	--ISO 3200   ;
```

## Organization of the code

Main functions are in the root folder. There are folders for `dataloader` objects,
`models` and command line `options` (though some options might be added in the 
main function files).
```
.
├── libBridge.cpp                        # Python wrapper around C code
├── library.py                           # Frequently used functions and wrappers
├── README.md                            # this README file
├── train.py
├── test.py
├── ...
├── 3rd-party
│   ├── imscript                         # image processing tools for linux
│   └── tvl1flow                         # TVL1 optical flow (in C)
├── data                                 # dataloaders
│   ├── __init__.py
│   ├── base_dataset.py
│   ├── axel4rec_dataset.py
│   ├── infer4rec_dataset.py
│   └── ...
├── models                               # models
│   ├── __init__.py
│   ├── base_model.py
│   ├── recurrentFeat_model.py
│   └── ...
├── networks                               # networks
│   ├── __init__.py
│   ├── emvd.py
│   └── ...

├── options                              # command line options
│   ├── __init__.py
│   ├── base_options.py                  # base options
│   ├── test_options.py                  # test options
│   └── train_options.py                 # train options
└── util                                 # utilities
    ├── __init__.py
    ├── flow_utils.py                    # compute flow and warp
    ├── get_data.py
    ├── util.py
    └── ...
```



## About

This source code is based on the following repo:
[https://github.com/VSAnimator/stgan](https://github.com/VSAnimator/stgan)

The following libraries are also included as part of the code:
* For computing the optical flow: [the IPOL implementation](http://www.ipol.im/pub/art/2013/26/) of the [TV-L1 optical flow method of Zack et al.](https://link.springer.com/chapter/10.1007/978-3-540-74936-3_22).
* For image I/O: [Enric Meinhardt's iio](https://github.com/mnhrdt/iio).
* For basic image manipulation: a reduced version of [Enric Meinhardt's imscript](https://github.com/mnhrdt/imscript).


