import os
import os.path
import numpy as np
import numpy.random
from copy import deepcopy
import torch
import torch.utils.data as udata
import random
from util.util import mkdir, mkdirs
import random
from data.base_dataset import BaseDataset, get_params, get_transform
from data.image_folder import make_dataset
from PIL import Image

from library import *

class axel4recDataset(BaseDataset):
    """A dataset class for training! It will provide random 3d patches based on axel's way of storing data.
    """
    @staticmethod
    def modify_commandline_options(parser, is_train=True):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.
        """
        BaseDataset.modify_commandline_options(parser, is_train)
        parser.add_argument('--warpeddata', action='store_true', default=False, help='Force this dataloader to give warped data too.')
        return parser

    def __init__(self, opt):
        """Initialize this dataset class.

        Parameters:
            opt (Option class) -- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        BaseDataset.__init__(self, opt)

        self.T, self.iT = define_transforms(opt)
        self.opt = opt
        self.rootdir = opt.dataroot

        self.gt_paths = os.path.join(self.rootdir, opt.gtFolder)
        self.n_paths = os.path.join(self.rootdir, opt.nFolder)

        # if opt.rawFolder in ["\noraw", " ", None]:
        #     self.israw = False
        #     self.ftype = '8bits'
        # else:
        self.israw = True
        assert opt.input_nc == 4, "The the input should be 4 channels!!!"
        # self.ftype = '10bits'
        self.ftype = '12bits'
        self.raw_paths = os.path.join(self.rootdir, opt.rawFolder)
        # self.createInputDataFromRaw()
        
        self.lbridge = CPPbridge('./build/libBridge.so')
        # self.w_paths = os.path.join(self.rootdir, opt.wFolder, opt.nFolder, opt.warp_method, opt.model_input_gen)
        # self.m_paths = os.path.join(self.rootdir, opt.mFolder, opt.nFolder, opt.warp_method, opt.model_input_gen)
        self.flow_paths = os.path.join(self.rootdir, opt.flowFolder, opt.nFolder, opt.warp_method, opt.model_input_gen)
        # mkdir(self.w_paths)
        mkdir(self.flow_paths)
        # mkdir(self.m_paths)


        assert(self.opt.load_size >= self.opt.patch_width)   # patch_width should be smaller than the size of loaded image
        self.input_nc = self.opt.input_nc
        self.output_nc = self.opt.output_nc
        
        gt_paths_list = os.scandir(self.gt_paths)
        if opt.videos is None:
            gt_paths_list = [c.path for c in gt_paths_list if not c.name.startswith('.') and c.is_dir()]
        else:
            opt.videos = opt.videos.split(',')
            gt_paths_list = [c.path for c in gt_paths_list if not c.name.startswith('.') and c.is_dir() and c.name in opt.videos]
        gt_paths_list = sorted(gt_paths_list)
        noise_paths_list = os.scandir(self.n_paths)
        if opt.videos is None:
            noise_paths_list = [c.path for c in noise_paths_list if not c.name.startswith('.') and c.is_dir()]
        else:
            noise_paths_list = [c.path for c in noise_paths_list if not c.name.startswith('.') and c.is_dir() and c.name in opt.videos]
        noise_paths_list = sorted(noise_paths_list)

        assert (len(gt_paths_list) == len(noise_paths_list))
        
        # warped_paths_list = [os.path.join(self.w_paths, os.path.basename(d)) for d in gt_paths_list ]
        # mkdirs(warped_paths_list)
        

        print ('%d videos' % len(gt_paths_list))
        self.gt_paths_list = gt_paths_list
        self.noise_paths_list = noise_paths_list

        self.total_depth = opt.patch_depth          # 5
        # This only works for axel4rec dataset, that creates opt.model_patch_depth
        self.patch_depth = opt.model_patch_depth       # 2 

        self.patch_width = opt.patch_width          # 96
        self.patch_stride = opt.patch_stride        # 3
        
        # self.createWarpedInputData()
        self.prepare_epoch()

    def prepare_epoch(self):
        """
        Load and randomize data
        """
        self.videos_noisy = []
        self.videos_gt = []
        self.videos_warped = []
        self.videos_flow = []
        self.videos_mask = []
        self.videos_noisy_path = []
        self.videos_gt_path = []

        num_frames_per_video = self.opt.frames2load

        for (video_path, video2_path) in zip(self.gt_paths_list, self.noise_paths_list):
            img_paths = list_video_files_at_dir(video_path)
            img2_paths = list_video_files_at_dir(video2_path)
            assert (len(img_paths) == len(img2_paths))

            num_frames = len(img_paths)  # 90
            start_frame = numpy.random.randint(num_frames-num_frames_per_video+1)
            img_paths = img_paths[start_frame:(start_frame+num_frames_per_video)]
            img2_paths = img2_paths[start_frame:(start_frame+num_frames_per_video)]

            video_gt = np.asarray([load_image(path,self.ftype) for path in img_paths], dtype=np.float32)
            video_noisy = np.asarray([load_image(path,self.ftype) for path in img2_paths], dtype=np.float32)
            self.videos_noisy_path.append(img2_paths)
            self.videos_gt_path.append(img_paths)
            self.videos_noisy.append(video_noisy)
            self.videos_gt.append(video_gt)


        for i, (video_path, video2_path) in enumerate( zip(self.gt_paths_list, self.noise_paths_list) ):
            self.videos_flow.append( [self.getFlowInputData(i,z) for z in range(num_frames_per_video-self.patch_depth+1)] )
        self.videos_flow = np.array(self.videos_flow).astype(np.float32)

        # if self.opt.warpeddata:
        #     optmod = deepcopy(self.opt)
        #     optmod.patch_depth = optmod.patch_depth - 1
        #     self.Twarped, self.iTwarped = define_transforms(optmod)
        #     for i, (video_path, video2_path) in enumerate( zip(self.gt_paths_list, self.noise_paths_list) ):
        #         self.videos_warped.append( [self.getWarpInputData(i,z) for z in range(num_frames_per_video-self.patch_depth+1)] )
        #     self.videos_warped = np.array(self.videos_warped).astype(np.float32)

        # if self.opt.mask:
        #     for i, (video_path, video2_path) in enumerate( zip(self.gt_paths_list, self.noise_paths_list) ):
        #         self.videos_mask.append( [self.getMaskInputData(i,z) for z in range(num_frames_per_video-self.patch_depth+1)] )
        #     self.videos_mask = np.array(self.videos_mask).astype(np.bool)
        # print(self.videos_gt_path)

        i = 0
        self.keys = []
        for v in self.videos_gt:
            zs = range(0, v.shape[0]-self.total_depth+1, self.patch_stride)
            ys = range(self.patch_width, v.shape[1]+1, self.patch_stride)
            xs = range(self.patch_width, v.shape[2]+1, self.patch_stride)
            xx, yy, zz = np.meshgrid(xs, ys, zs)
            xx = np.asarray(xx.flatten(), dtype=np.uint32)
            yy = np.asarray(yy.flatten(), dtype=np.uint32)
            zz = np.asarray(zz.flatten(), dtype=np.uint32)
            self.keys.append(np.stack([i*np.ones([len(xx)], dtype=np.uint32), xx, yy, zz]).T)
            i = i + 1

        self.keys = np.concatenate(self.keys, axis=0)
        self.num_keys = self.keys.shape[0]
        self.indices = [i for i in range(self.num_keys)]

        random.shuffle(self.indices)


    def __len__(self):
        return self.num_keys

    def data_num_channels(self):
        return 3

    def __getitem__(self, index):
        """Return a data point and its metadata information.

        Parameters:
            index - - a random integer for data indexing

        Returns a dictionary that contains gt, n, gt_path, n_path
            gt (tensor) - - Groundtruth 3d patch
            n (tensor) - - Noisy 3d patch
            gt_path (str) - - Image path associated to the last 2d patch in gt
            n_path (str) - - Image path associated to the last 2d patch in n
        """

        key = self.keys[self.indices[index],:]
        patch_width = self.patch_width
        i = key[0]
        x = key[1]
        y = key[2]
        z = key[3]

        gt = self.videos_gt[i][z:(z+self.total_depth), (y-patch_width):y, (x-patch_width):x,:]
        noise = self.videos_noisy[i][z:(z+self.total_depth), (y-patch_width):y, (x-patch_width):x,:]        

        gt = gt.transpose(0, 3, 1, 2)
        gt = gt.reshape([gt.shape[0] * gt.shape[1], gt.shape[2], gt.shape[3]])
        gt = gt.transpose(1, 2, 0)

        noise = noise.transpose(0, 3, 1, 2)
        noise = noise.reshape([noise.shape[0] * noise.shape[1], noise.shape[2], noise.shape[3]])
        noise = noise.transpose(1, 2, 0)

        flows = self.videos_flow[i,z:(z+self.total_depth-self.patch_depth+1),:,(y-patch_width):y, (x-patch_width):x, :]
        flows = flows.transpose(0, 1, 4, 2, 3)
        flows = torch.from_numpy(flows)

        # if self.opt.mask:
        #     mask = self.videos_mask[i,z:(z+self.total_depth-self.patch_depth+1),:,(y-patch_width):y, (x-patch_width):x, :]
        #     mask = mask.transpose(0, 1, 4, 2, 3).astype(np.float32)
        #     mask = torch.from_numpy(mask)
        #     return {'mask': mask, 'gt': self.T(gt), 'n': self.T(noise), 'flow': flows, 'gt_path': self.videos_gt_path[i][z+self.total_depth-1], 'n_path': self.videos_noisy_path[i][z+self.total_depth-1]}

        # if self.opt.warpeddata:
        #     warps = self.videos_warped[i,z:(z+self.total_depth-self.patch_depth+1),:,(y-patch_width):y, (x-patch_width):x, :]
        #     warps = warps.transpose(0,1,4,2,3)
        #     warps = warps.reshape([warps.shape[0] * warps.shape[1] * warps.shape[2], warps.shape[3], warps.shape[4]])
        #     warps = warps.transpose(1, 2, 0)
        #     return {'warped': self.Twarped(warps), 'gt': self.T(gt), 'n': self.T(noise), 'flow': flows, 'gt_path': self.videos_gt_path[i][z+self.total_depth-1], 'n_path': self.videos_noisy_path[i][z+self.total_depth-1]}

        return {'gt': self.T(gt), 'n': self.T(noise), 'flow': flows, 'gt_path': self.videos_gt_path[i][z+self.total_depth-1], 'n_path': self.videos_noisy_path[i][z+self.total_depth-1]}
        return {'gt': gt, 'n': self.T(noise), 'flow': flows, 'gt_path': self.videos_gt_path[i][z+self.total_depth-1], 'n_path': self.videos_noisy_path[i][z+self.total_depth-1]}
        
