"""This module implements an abstract base class (ABC) 'BaseDataset' for datasets.

It also includes common transformation functions (e.g., get_transform, __scale_width), which can be later used in subclasses.
"""
import random
import numpy as np
import torch.utils.data as data
from PIL import Image
import torchvision.transforms as transforms
from abc import ABC, abstractmethod

from util.addNoise import addNoise
from os import scandir
from os.path import join, isfile, basename, splitext
from library import iio_write, list_video_files_at_dir, iio_read, warpedimagefile, pathdiff, load_image
from util.util import mkdir, mkdirs
from util.flow_utils import single_warp_noTorch, single_warp_Torch
import time

class BaseDataset(data.Dataset, ABC):
    """This class is an abstract base class (ABC) for datasets.

    To create a subclass, you need to implement the following four functions:
    -- <__init__>:                      initialize the class, first call BaseDataset.__init__(self, opt).
    -- <__len__>:                       return the size of dataset.
    -- <__getitem__>:                   get a data point.
    -- <modify_commandline_options>:    (optionally) add dataset-specific options and set default options.
    """

    def __init__(self, opt):
        """Initialize the class; save the options in the class

        Parameters:
            opt (Option class)-- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        self.opt = opt
        self.root = opt.dataroot

    @staticmethod
    def modify_commandline_options(parser, is_train):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.
        """
        parser.add_argument('--frames2load', type=int, default=10, help='Number of frames per video to load to RAM.')
        parser.add_argument('--ISOgen', type=int, default=3500, help='If noisy data is to be generated, it will be generated with this ISO value.')
        return parser

    @abstractmethod
    def __len__(self):
        """Return the total number of images in the dataset."""
        return 0

    @abstractmethod
    def __getitem__(self, index):
        """Return a data point and its metadata information.

        Parameters:
            index - - a random integer for data indexing

        Returns:
            a dictionary of data with their names. It ususally contains the data itself and its metadata information.
        """
        pass

    @abstractmethod
    def prepare_epoch(self):
        """Load and randomize data"""
        print("nothing to do in prepare_epoch")
        pass

    def getWarpInputData(self, idx_video, start_frame):
        """ This function returns the warped data needed
        to match the reference one (the last frame). 
        """
        i = idx_video
        z = start_frame

        toCode = splitext( basename(self.videos_noisy_path[i][z+self.patch_depth-1]) )[0]
        wfolder = join(self.w_paths, pathdiff(self.videos_noisy_path[i][z+self.patch_depth-1], self.n_paths))
        mkdir(wfolder)
        videos_w = []
        for n in range(self.patch_depth-1):
            fromCode = splitext( basename(self.videos_noisy_path[i][z+n]) )[0]
            wimagefile = warpedimagefile(wfolder, fromCode, toCode)
            if isfile(wimagefile):
                warped = load_image(wimagefile,self.ftype)
            else:
                print('Warped image not found. Exiting...')
                exit()
            videos_w.append( warped )
        return np.array(videos_w).astype(np.float32)

    def getMaskInputData(self, idx_video, start_frame):
        """ This function returns all masks needed
        with respect to the reference frame. 
        """
        i = idx_video
        z = start_frame
        videos_mask = []
        toCode = splitext( basename(self.videos_noisy_path[i][z+self.patch_depth-1]) )[0]
        mfolder = join(self.m_paths, pathdiff(self.videos_noisy_path[i][z+self.patch_depth-1], self.n_paths))
        for n in range(self.patch_depth-1):
            fromCode = splitext( basename(self.videos_noisy_path[i][z+n]) )[0]
            mimagefile = warpedimagefile(mfolder, fromCode, toCode)
            if isfile(mimagefile):
                mask = iio_read(mimagefile).astype(np.float32)
            else:
                print("Mask not found in function getMaskInputData")
                exit()
            videos_mask.append(mask)
        return np.array(videos_mask).astype(np.bool)

    def getFlowInputData(self, idx_video, start_frame):
        """ This function return the flow data needed
        to match the reference frame (the last frame). 
        """
        i = idx_video
        z = start_frame
        videos_flow = []
        toCode = splitext( basename(self.videos_noisy_path[i][z+self.patch_depth-1]) )[0]
        flowfolder = join(self.flow_paths, pathdiff(self.videos_noisy_path[i][z+self.patch_depth-1], self.n_paths))
        for n in range(self.patch_depth-1):
            fromCode = splitext( basename(self.videos_noisy_path[i][z+n]) )[0]
            flowimagefile = warpedimagefile(flowfolder, fromCode, toCode)
            if isfile(flowimagefile):
                flow = iio_read(flowimagefile).astype(np.float32)
            else:
                print("flow not found in function getFlowInputData")
                exit()
            videos_flow.append( flow )   
        return np.array(videos_flow).astype(np.float32)

    def createWarpedInputData(self):
        """ This function creates (if not existing) all warped data needed
        to match the reference frame (the last frame). 
        """
        if self.opt.data_nocheck:
            return
        start_time = time.time()
        for video_path, video2_path in zip(self.gt_paths_list, self.noise_paths_list):
            img2_paths = list_video_files_at_dir(video2_path)
            if self.opt.model_input_gen == 'lastnoisy':
                img_paths = list_video_files_at_dir(video_path)
            for z in range(len(img2_paths)-self.patch_depth+1):
                toCode = splitext( basename(img2_paths[z+self.patch_depth-1]) )[0]
                wfolder = join(self.w_paths, pathdiff(img2_paths[z+self.patch_depth-1], self.n_paths))
                flowfolder = join(self.flow_paths, pathdiff(img2_paths[z+self.patch_depth-1], self.n_paths))
                mfolder = join(self.m_paths, pathdiff(img2_paths[z+self.patch_depth-1], self.n_paths))
                mkdir(wfolder)
                mkdir(flowfolder)
                mkdir(mfolder)
                img2 = iio_read(img2_paths[z+self.patch_depth-1]).astype(np.float32)
                for n in range(self.patch_depth-1):
                    fromCode = splitext( basename(img2_paths[z+n]) )[0]
                    wimagefile = warpedimagefile(wfolder, fromCode, toCode)            
                    mimagefile = warpedimagefile(mfolder, fromCode, toCode)
                    flowimagefile = warpedimagefile(flowfolder, fromCode, toCode)
                    if not isfile(wimagefile) or not isfile(flowimagefile):
                        if self.opt.model_input_gen == 'lastnoisy':
                            img1 = iio_read(img_paths[z+n]).astype(np.float32)
                        else:
                            img1 = iio_read(img2_paths[z+n]).astype(np.float32)
                        if isfile(flowimagefile):
                            flow = iio_read(flowimagefile).astype(np.float32)
                            if not self.opt.warp_in_torch:
                                warped = single_warp_noTorch(img1, flow)
                            else:
                                warped = single_warp_Torch(img1, flow)
                        else:
                            warped, flow = self.lbridge.WarpWithBackwardFlow(img1, img2,
                                    flowtype=self.opt.warp_method,
                                    noTorch=not self.opt.warp_in_torch,
                                    giveMeFlow=True)

                            iio_write(flow.astype(np.float32), flowimagefile)
                        iio_write(warped.astype(np.float32), wimagefile)                        
                    elif not isfile(mimagefile):
                        warped = iio_read(wimagefile)
                    if self.opt.mask and not isfile(mimagefile):
                        mask = self.lbridge.compute_mask(img2, warped)
                        iio_write(mask.astype(np.bool), mimagefile)
        print('Warp-Flow-Mask creation/checking: %d sec' % (time.time() - start_time))
    
    def createInputDataFromRaw(self):
        """ This function creates (if not existing) both groundtruth 
        and noisy data from raw data.
        """
        print('==================================================')
        if self.opt.data_nocheck:
            return
        start_time = time.time()
        video_paths_list = scandir(self.raw_paths)
        video_paths_list = [c.path for c in video_paths_list if not c.name.startswith('.') and c.is_dir()]
        
        gt_paths_list = [join(self.gt_paths, basename(d)) for d in video_paths_list ]
        noise_paths_list = [join(self.n_paths, basename(d)) for d in video_paths_list ]
        mkdirs(gt_paths_list)
        mkdirs(noise_paths_list)

        for video_path in video_paths_list:
            videoid = basename(video_path)
            raw_paths = list_video_files_at_dir(video_path)
            for filename in raw_paths:
                name = splitext( basename(filename) )[0]
                gtpath = join(self.gt_paths, videoid, name+'.tif')
                npath = join(self.n_paths, videoid, name+'.tif')
                if (not isfile(gtpath)) or (not isfile(npath)):
                    raw_img = np.single(np.fromfile(filename,np.uint16).reshape((720, 1280)))
                    bayer_img = bayer2channels(raw_img, pattern='GRBG')
                    iio_write(bayer_img.astype(np.uint16), gtpath)
                    n_img = addNoise(raw_img, ISO=self.opt.ISOgen)
                    nbayer_img = bayer2channels(n_img.astype(np.uint16), pattern='GRBG')
                    iio_write(nbayer_img, npath)
        print('Data From RAW creation/checking: %d sec' % (time.time() - start_time))

def get_params(opt, size):
    w, h = size
    new_h = h
    new_w = w
    if opt.preprocess == 'resize_and_crop':
        new_h = new_w = opt.load_size
    elif opt.preprocess == 'scale_width_and_crop':
        new_w = opt.load_size
        new_h = opt.load_size * h // w

    x = random.randint(0, np.maximum(0, new_w - opt.patch_width))
    y = random.randint(0, np.maximum(0, new_h - opt.patch_width))

    flip = random.random() > 0.5

    return {'crop_pos': (x, y), 'flip': flip}


def get_transform(opt, params=None, grayscale=False, method=Image.BICUBIC, convert=True):
    transform_list = []
    if grayscale:
        transform_list.append(transforms.Grayscale(1))
    if 'resize' in opt.preprocess:
        osize = [opt.load_size, opt.load_size]
        transform_list.append(transforms.Resize(osize, method))
    elif 'scale_width' in opt.preprocess:
        transform_list.append(transforms.Lambda(lambda img: __scale_width(img, opt.load_size, method)))

    if 'crop' in opt.preprocess:
        if params is None:
            transform_list.append(transforms.RandomCrop(opt.patch_width))
        else:
            transform_list.append(transforms.Lambda(lambda img: __crop(img, params['crop_pos'], opt.patch_width)))

    if opt.preprocess == 'none':
        transform_list.append(transforms.Lambda(lambda img: __make_power_2(img, base=4, method=method)))

    if not opt.no_flip:
        if params is None:
            transform_list.append(transforms.RandomHorizontalFlip())
        elif params['flip']:
            transform_list.append(transforms.Lambda(lambda img: __flip(img, params['flip'])))

    if convert:
        transform_list += [transforms.ToTensor()]
        if grayscale:
            transform_list += [transforms.Normalize((0.5),(0.5))]
        else:
            transform_list += [transforms.Normalize((0.5, 0.5, 0.5),
                                                (0.5, 0.5, 0.5))]
    return transforms.Compose(transform_list)


def __make_power_2(img, base, method=Image.BICUBIC):
    ow, oh = img.size
    h = int(round(oh / base) * base)
    w = int(round(ow / base) * base)
    if (h == oh) and (w == ow):
        return img

    __print_size_warning(ow, oh, w, h)
    return img.resize((w, h), method)


def __scale_width(img, target_width, method=Image.BICUBIC):
    ow, oh = img.size
    if (ow == target_width):
        return img
    w = target_width
    h = int(target_width * oh / ow)
    return img.resize((w, h), method)


def __crop(img, pos, size):
    ow, oh = img.size
    x1, y1 = pos
    tw = th = size
    if (ow > tw or oh > th):
        return img.crop((x1, y1, x1 + tw, y1 + th))
    return img


def __flip(img, flip):
    if flip:
        return img.transpose(Image.FLIP_LEFT_RIGHT)
    return img


def __print_size_warning(ow, oh, w, h):
    """Print warning information about image size(only print once)"""
    if not hasattr(__print_size_warning, 'has_printed'):
        print("The image size needs to be a multiple of 4. "
              "The loaded image size was (%d, %d), so it was adjusted to "
              "(%d, %d). This adjustment will be done to all images "
              "whose sizes are not multiples of 4" % (ow, oh, w, h))
        __print_size_warning.has_printed = True
