"""General-purpose training script for video denoising
"""
import time
from options.train_options import TrainOptions
from data import create_dataset
from models import create_model
from util.visualizer import Visualizer, save_images
import torch
import copy
from os.path import isfile, dirname, basename, join
from library import load_ordereddict, save_ordereddict, pathdiff


def resume_training(model, opt, statusfile):
    if opt.autoresume and isfile(statusfile):
        status = load_ordereddict( statusfile )
        epoch = status['epoch']
        epoch_to_load = epoch if True else 'latest'
        model.load_networks(epoch_to_load)
        model.update_learning_rate(epoch)
        opt.epoch_count = epoch + 1
    else:
        if opt.autoresume:
            print("---> No status file for resuming!!!")
        status = {'epoch': opt.epoch_count}
    return status        

if __name__ == '__main__':
    """
    Simplifying the code. The spirit is to keep the stuff related to the emvd/recurrentFeat network(model).
                          Remove the unnecessary parts. Remove unnecessary generalizations.

    The main idea is to follow the logic of this train.py file, check what kind of functions are needed.
    1. Options
    2. DataLoader, including train_dataset and validation_dataset, they use different dataloaders.
                    Functions: train_dataset.prepare_epoch() 
                    Attibutes: val_dataset.dataset.n_paths
    3. Model. create_model().  Functions: model.setup() .set_input() .optimize_paras() .compute_visuals() .compute_losses() .get_current_losses/visuals()
    4. Visualizer()  This is a class for visualizing images. Seems having connection with html, which won't be used. chech this later

    """
    # start with the model. After modification, check if we can create correct model, and if all functions function well.

    
    opt = TrainOptions().parse()   # get training options
    train_dataset = create_dataset(opt)  # create a dataset given opt.dataset_mode and other options
    train_dataset_size = len(train_dataset)    # get the number of images in the dataset.
    print('The number of training images = %d' % train_dataset_size)

    if not opt.no_val:
        from library import define_transforms
        opt_val = copy.deepcopy(opt)
        opt_val.patch_depth = 1
        singleT, singleiT = define_transforms(opt_val)
        opt_val = copy.deepcopy(opt)
        opt_val.dataroot, opt_val.dataset_mode, opt_val.max_dataset_size, opt_val.phase, opt_val.videos = opt.val_dataroot, opt.val_dataset_mode, opt.val_max_dataset_size, 'val', opt.val_videos
        opt_val.num_threads = 0   # test code only supports num_threads = 1
        opt_val.batch_size = 1    # test code only supports batch_size = 1
        opt_val.serial_batches = True  # disable data shuffling; comment this line if results on randomly chosen images are needed.
        opt_val.no_flip = True    # no flip; comment this line if results on flipped images are needed.
        # opt_val.display_id = -1   # no visdom display; the test code saves the results to a HTML file.
        if hasattr(opt, 'model_patch_depth'):
            opt_val.patch_depth = opt.model_patch_depth
        val_dataset = create_dataset(opt_val)
        # define the website directory
        save_dir = join(opt.checkpoints_dir, opt.name, "val_visuals", "images")
        # webpage = html.HTML(web_dir, 'Experiment = %s, Phase = %s, Epoch = %s' % \
        #                 (opt.name, "validation", "val"))
        print('The number of validation images = %d' % len(val_dataset))

    torch.backends.cudnn.benchmark = True  # CUDNN optimization

    model = create_model(opt)      # create a model given opt.model and other options
    model.setup(opt)               # regular setup: load and print networks; create schedulers
    if opt.visuals == "none":
        opt.display_id = -1
        opt.no_html = True
    visualizer = Visualizer(opt)   # create a visualizer that display/save images and plots
    total_iters = 0                # the total number of training iterations

    statusfile = model.save_dir + "/status.pkl"
    status = resume_training(model, opt, statusfile)

    # save initialization
    if opt.epoch_count == 1:
        model.save_networks('0')


    for epoch in range(opt.epoch_count, opt.niter + opt.niter_decay + 1):    # outer loop for different epochs; we save the model by <epoch_count>, <epoch_count>+<save_latest_freq>
        model.train() # Pre-training step

        epoch_start_time = time.time()  # timer for entire epoch
        iter_data_time = time.time()    # timer for data loading per iteration
        epoch_iter = 0                  # the number of training iterations in current epoch, reset to 0 every epoch


        for i, data in enumerate(train_dataset):  # inner loop within one epoch
            iter_start_time = time.time()  # timer for computation per iteration

            # information about training status
            data['epoch'] = epoch
            data['epoch_length'] = len(train_dataset)
            data['epoch_iter'] = i

            if total_iters % opt.print_freq == 0:
                t_data = iter_start_time - iter_data_time
            visualizer.reset()
            total_iters += opt.batch_size
            epoch_iter += opt.batch_size
            model.set_input(data)         # unpack data from dataset and apply preprocessing
            model.optimize_parameters()   # calculate loss functions, get gradients, update network weights

            if not opt.visuals == "none" and total_iters % opt.display_freq == 0:   # display images on visdom and save images to a HTML file
                save_result = total_iters % opt.update_freq == 0
                model.compute_visuals()
                visualizer.display_current_results(model.get_current_visuals(), epoch, save_result)

            if total_iters % opt.print_freq == 0:    # print training losses and save logging information to the disk
                losses = model.get_current_losses()
                t_comp = (time.time() - iter_start_time) / opt.batch_size
                visualizer.print_current_losses(epoch, epoch_iter, losses, t_comp, t_data)
                # if opt.display_id > 0:
                    # visualizer.plot_current_losses(epoch, float(epoch_iter) / train_dataset_size, losses)

            if total_iters % opt.save_latest_freq == 0:   # cache our latest model every <save_latest_freq> iterations
                print('saving the latest model (epoch %d, total_iters %d)' % (epoch, total_iters))
                save_suffix = 'iter_%d' % total_iters if opt.save_by_iter else 'latest'
                model.save_networks(save_suffix)

            iter_data_time = time.time()
        if epoch % opt.save_epoch_freq == 0:              # cache our model every <save_epoch_freq> epochs
            print('saving the model at the end of epoch %d, iters %d' % (epoch, total_iters))
            model.save_networks('latest')
            model.save_networks(epoch)
            status['epoch'] = epoch
            save_ordereddict(status, statusfile)

        train_dataset.prepare_epoch() # Rerandomize training set
        
        if not opt.no_val and epoch % opt.val_epoch_freq == 0:
            print("Compute the validation at epoch %d!" % (epoch))
            model.isTrain = False
            # bak_visual_names = model.visual_names
            model.set_visuals(opt.val_visuals)
            # model.visual_names = ['denoised', 'fusion'] # FIXME: fusion only works for EMVD network
            # model.visual_names = ['fusion', 'denoised', 'weights', 'n_map', 'f_map']  # 
            model.eval() # Correctly set the BN layers before inference
            val_start_time = time.time()
            val_losses = model.get_current_losses().copy()
            for k, _ in val_losses.items():
                val_losses[k] = 0.0
            with torch.no_grad():
                    lastvideopath = ''
                    for i, data in enumerate(val_dataset):
                            thisvideopath = dirname(data['gt_path'][0])
                            data['FirstOfVideo'] = not thisvideopath == lastvideopath
                            model.set_input(data)
                            model.test()
                            model.compute_losses()

                            visuals = model.get_current_visuals()  # get image results
                            img_path = model.get_image_paths()     # get image paths
                            if i % 5 == 0:  # save images to an HTML file
                                print('processing (%04d)-th image... %s' % (i, img_path))

                            sfolder = pathdiff(img_path[0], val_dataset.dataset.n_paths)
                            save_images(save_dir, visuals, [basename(img_path[0])],
                                        subfolder=sfolder, iT=singleiT)

                            losses = model.get_current_losses()
                            for k, v in losses.items():
                                val_losses[k] += v
                            lastvideopath = thisvideopath
                    for k in val_losses.keys():
                        val_losses[k] /= len(val_dataset)
            message = "["
            for k, v in val_losses.items():
                message += '%s: %.3f, ' % (k, v)
            message = message[:-2] + "]"
            val_losses = dict([(k+"_valLoss",v) for k,v in val_losses.items()])        
            val_losses['lr'] = model.optimizers[0].param_groups[0]['lr']
            
            if val_losses['Denoiser_valLoss'] < model.best_val_score:
                model.save_networks('latest_val')
                model.best_val_score = val_losses['Denoiser_valLoss']

            visualizer.print_current_val_losses(epoch, val_losses, time.time()-val_start_time, len(val_dataset))
            # if opt.display_id > 0:
            #     visualizer.plot_current_val_losses(epoch, 1.0 , val_losses)
            model.set_visuals(opt.visuals)
            model.isTrain = True

        print('End of epoch %d / %d \t Time Taken: %d sec' % (epoch, opt.niter + opt.niter_decay, time.time() - epoch_start_time))
        model.update_learning_rate(epoch)    # update learning rates at the end of every epoch.
