# -*- coding: utf-8 -*-
"""
Created on Fri Mar 29 15:01:48 2019

@author: m80054475
"""
import numpy as np
import scipy.ndimage   


def bayer2channels(In, pattern='GRBG'):
    # bayer2channels returns a 4-channel image from the Bayer CFA grayscale image
    #
    # GrRBGb = bayer2channels(I)
    # GrRBGb = bayer2channels(I, pattern)
    #
    # Pattern string can be RGGB, BGGR, GBRG, or GRBG (default if omitted).

    M, N = np.shape(In)

    if (str.lower(pattern) == 'rggb'):
        k = 0
    elif (str.lower(pattern) == 'bggr'):
        k = 2
    elif (str.lower(pattern) == 'gbrg'):
        k = 1
    else:
        k = -1

    mask_r = np.tile(np.rot90(np.array([[1, 0], [0, 0]]), k), [
                     np.int(M/2), np.int(N/2)]).astype(bool)
    mask_gr = np.tile(np.rot90(np.array([[0, 1], [0, 0]]), k), [
                      np.int(M/2), np.int(N/2)]).astype(bool)
    mask_gb = np.tile(np.rot90(np.array([[0, 0], [1, 0]]), k), [
                      np.int(M/2), np.int(N/2)]).astype(bool)
    mask_b = np.tile(np.rot90(np.array([[0, 0], [0, 1]]), k), [
                     np.int(M/2), np.int(N/2)]).astype(bool)

    B = In[mask_b].reshape(np.int(M/2), np.int(N/2), 1)
    R = In[mask_r].reshape(np.int(M/2), np.int(N/2), 1)
    Gb = In[mask_gb].reshape(np.int(M/2), np.int(N/2), 1)
    Gr = In[mask_gr].reshape(np.int(M/2), np.int(N/2), 1)

    output = np.concatenate((Gr, R, B, Gb), axis=2)

    return output


def channels2bayer(In, pattern='GRBG'):
    # channels2bayer returns a Bayer CFA grayscale image from the 4-channel image
    #
    # BayerImg = channels2bayer( GrRBGb )
    # BayerImg = channels2bayer( GrRBGb, pattern)
    #
    # Pattern string can be RGGB, BGGR, GBRG, or GRBG (default if omitted).

    M, N = 2*np.array(np.shape(In)[:2])

    if (str.lower(pattern) == 'rggb'):
        k = 0
    elif (str.lower(pattern) == 'bggr'):
        k = 2
    elif (str.lower(pattern) == 'gbrg'):
        k = 1
    else:
        k = -1

    mask_r = np.tile(np.rot90(np.array([[1, 0], [0, 0]]), k), [
                     np.int(M/2), np.int(N/2)]).astype(bool)
    mask_gr = np.tile(np.rot90(np.array([[0, 1], [0, 0]]), k), [
                      np.int(M/2), np.int(N/2)]).astype(bool)
    mask_gb = np.tile(np.rot90(np.array([[0, 0], [1, 0]]), k), [
                      np.int(M/2), np.int(N/2)]).astype(bool)
    mask_b = np.tile(np.rot90(np.array([[0, 0], [0, 1]]), k), [
                     np.int(M/2), np.int(N/2)]).astype(bool)

    Gr, R, B, Gb = In[:, :, 0], In[:, :, 1], In[:, :, 2], In[:, :, 3]

    output = np.zeros(shape=(M, N), dtype=In.dtype)
    output[mask_r] = R.flatten()
    output[mask_b] = B.flatten()
    output[mask_gb] = Gb.flatten()
    output[mask_gr] = Gr.flatten()

    return output

def demmhc (In, pattern='GRBG'):
    # DEMMHC Interpolates Bayer CFA grayscale image to full RGB image
    #
    # RGB = demmhc(I)
    # RGB = demmhc(I, pattern)
    #
    # Pattern string can be RGGB, BGGR, GBRG, or GRBG (default if omitted).
    #
    # Based on Malvar-He-Cutler gradient-corrected bilinear interpolation method.
    
    M, N = np.shape(In)

    if (str.lower(pattern)=='rggb'):
        k = 0
    elif (str.lower(pattern)=='bggr'):
        k = 2
    elif (str.lower(pattern)=='gbrg'):
        k = 1
    else :
        k = -1

    mask_r  = np.tile(np.rot90(np.array([[1, 0],[0, 0]]), k), [np.int(M/2), np.int(N/2)]).astype(bool)
    mask_gr = np.tile(np.rot90(np.array([[0, 1],[0, 0]]), k), [np.int(M/2), np.int(N/2)]).astype(bool)
    mask_gb = np.tile(np.rot90(np.array([[0, 0],[1, 0]]), k), [np.int(M/2), np.int(N/2)]).astype(bool)
    mask_b  = np.tile(np.rot90(np.array([[0, 0],[0, 1]]), k), [np.int(M/2), np.int(N/2)]).astype(bool)
    
    R = In*mask_r
    G = In*np.logical_or(mask_gr,mask_gb)
    B = In*mask_b
    
    #Calculation of horizontal and vertical gradient corrections from green
    f = np.array([[0,0,1,0,0],[0,-2,0,-2,0],[-2,8,10,8,-2],[0,-2,0,-2,0],[0,0,1,0,0]])/16
    Gh = scipy.ndimage.correlate(G, f, mode='constant')
    Gv = scipy.ndimage.correlate(G, f.transpose(), mode='constant')

    #Interpolation for the green at the missing points

    G = G + scipy.ndimage.correlate(G, np.array([[0,1,0],[1,0,1],[0,1,0]])/4, mode='constant') \
    +scipy.ndimage.correlate(R+B, np.array([[0,0,-1,0,0],[0,0,0,0,0],[-1,0,4,0,-1],[0,0,0,0,0],[0,0,-1,0,0]])/8,mode='constant')

    # Calculation of the missing blue/red pixels at the red/blue location
    RB = scipy.ndimage.correlate(R+B, np.array([[0,0,-3,0,0],[0,4,0,4,0],[-3,0,12,0,-3],[0,4,0,4,0],[0,0,-3,0,0]])/16,\
                            mode='constant')

    # Finally, interpolate the missing blue pixels, and add the correction terms at appropriate locations
    B = B + scipy.ndimage.correlate(B, np.array([[0,1,0],[1,0,1],[0,1,0]])/2, mode='constant') + \
        RB*mask_r + Gh*mask_gb + Gv*mask_gr

    # Finally, interpolate the missing red pixels, and add the correction terms at appropriate locations
    R = R + scipy.ndimage.correlate(R, np.array([[0,1,0],[1,0,1],[0,1,0]])/2, mode='constant') + \
        RB*mask_b + Gh*mask_gr + Gv*mask_gb

    RB=[]
    Gh=[]
    Gv=[]

    out = np.zeros((M,N,3))
    out[:,:,0]=R
    out[:,:,1]=G
    out[:,:,2]=B
    return out
